/* Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */

package org.apache.myfaces.portlet.faces.application;

import java.io.IOException;

import javax.faces.FacesException;
import javax.faces.application.Application;
import javax.faces.application.ViewHandler;
import javax.faces.application.ViewHandlerWrapper;
import javax.faces.component.UIComponent;
import javax.faces.component.UIViewRoot;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;

import javax.portlet.MimeResponse;
import javax.portlet.PortletContext;
import javax.portlet.faces.Bridge;
import javax.portlet.faces.BridgeUtil;
import javax.portlet.faces.annotation.PortletNamingContainer;
import javax.portlet.faces.component.PortletNamingContainerUIViewRoot;

import org.apache.myfaces.portlet.faces.bridge.BridgeImpl;

/**
 * View handler implementation for JSF portlet bridge.
 * 
 * The only method we override here is getActionURL().
 * 
 * TODO JSF 1.2 note: JSF 1.2 RI implements ViewHandler.renderView() differently in order to handle
 * emitting non-JSF markup that follows the JSF tags after the JSF renders correctly. Unfortunately,
 * the RI handles this by introducing several servlet dependencies. Currently, the bridge handles
 * this by overriding the renderView() and ignoring (not interleaving) the non-JSF markup - see HACK
 * below
 */
public class PortletViewHandlerImpl extends ViewHandlerWrapper
{

  // the ViewHandler to delegate to
  private ViewHandler mDelegate;
  private Bridge.BridgeRenderPolicy mRenderPolicy = null;
  
  public PortletViewHandlerImpl(ViewHandler handler)
  {
    mDelegate = handler;
  }
  
  @Override
  public ViewHandler getWrapped()
  {
    return mDelegate;
  }

  @Override
  public String getActionURL(FacesContext context, String viewId)
  {
    // action URLs are processed by the bridge in encodeActionURL
    // however the bridge extends Faces navigation rule support in that it
    // allows a to-view-id element to contain an EL expression.
    // We recognize this EL expression here and evaluate to a viewid
    // before delegating
    // Do nothing when not running in portlet request
    
    if (!BridgeUtil.isPortletRequest())
    {
      return super.getActionURL(context, viewId);
    }
    
    // else its a portlet request so processing accoring to bridge rules
    /* Note: later versions of glassfish jsf append a / if not in the nav rule
    * -- remove it if blocking the expression.
    */
    if (viewId.startsWith("/#"))
    {
      viewId = viewId.substring(1);
    }
    
    if (viewId.startsWith("#"))
    {
      // evaluate this as an EL expression
      viewId = (String) context.getApplication().evaluateExpressionGet(context, viewId, String.class);
      if (viewId == null)
      {
        //TODO:  at least log an error.
      }
    }
    
    // Faces can't do suffix mapping (extension mapping) properly if there is a query string
    int qsLoc = viewId.indexOf('?');
    if (qsLoc < 0) qsLoc = viewId.length();
    
    String actionURL = super.getActionURL(context, viewId.substring(0, qsLoc));

    // Now add the parameters back on (if needed)
    if (qsLoc < viewId.length())
    {
      int aLoc = actionURL.indexOf('?');
      if (aLoc < 0)
      {
        actionURL = actionURL.concat(viewId.substring(qsLoc));
      }
      else if (viewId.length() > (qsLoc + 1))
      {
        actionURL = actionURL + "&" + viewId.substring(qsLoc + 1);
      }
    }
    
    return actionURL;
    
  }

  @Override
  public UIViewRoot createView(FacesContext facesContext, String viewId)
  {
    // The Bridge extends Faces navigation rule support in that it
    // allows a to-view-id element to contain an EL expression.
    // We recognize this EL expression here and evaluate to a viewid
    // before delegating
    // Do nothing when not running in portlet request
    // TODO: Should some of this logic be moved to the ViewDeclarationLanguage?
    
    if (!BridgeUtil.isPortletRequest())
    {
      return super.createView(facesContext, viewId);
    }
    
    // else its a portlet request so processing accoring to bridge rules
    
    /* Note: later versions of glassfish jsf append a / if not in the nav rule
    * -- remove it if blocking the expression.
    */
    if (viewId.startsWith("/#"))
    {
      viewId = viewId.substring(1);
    }
    
    if (viewId.startsWith("#"))
    {
      // evaluate this as an EL expression
      viewId = (String) facesContext.getApplication().evaluateExpressionGet(facesContext, viewId, String.class);
      if (viewId == null)
      {
        //TODO:  at least log an error.
      }
    }
  
    UIViewRoot viewRoot = null;
    
    // The bridge allows viewIds to carry querystrings -- e.g. a qs that has the 
    // mode parameter.  Unfortunately, underlying Faces impls don't expect any query
    // string when mapping from the viewId to the view -- in particular they don't
    // account for the query string when the Faces servlet is suffix mapped (extension
    // mapped).  So strip the query string here -- create the view and add back
    int qsLoc = viewId.indexOf('?');
    if (qsLoc < 0)
      qsLoc = viewId.length();
  
    viewRoot = super.createView(facesContext, viewId.substring(0, qsLoc));
    
    // As we have provided an ApplicationImpl to override createComponent
    // to detect when to swap in our NamingContainer the returned viewRoot
    // from above should already be the one we want.  Because its possible
    // its still the native UIViewRoot -- i.e. createComponent wasn't called 
    // the following code checks/and does the right thing to satisfy the spec.
    
    if (viewRoot.getClass() == UIViewRoot.class &&
        UIViewRoot.class.isAnnotationPresent(PortletNamingContainer.class))
    {
      String prior = setUIViewRootComponent(facesContext);
      try 
      {
        viewRoot = super.createView(facesContext, viewId.substring(0, qsLoc));
      }
      finally
      {
        if (prior != null)
          resetUIViewRootComponent(facesContext, prior);
      }
    }
      
    if (qsLoc < viewId.length())
    {
      // Add the QueryString as an additional attribute to the UIViewRoot.
      // Note:  only expected situation is when a NavigationHandler returns a viewId with a QS
      // as we support this in the faces-config.xml to introduce modes, etc.
      // The Bridge (ActionHandler) will pull this when it constructs the target
      // of the actionResponse based on this current ViewId.
      viewRoot.getAttributes().put(BridgeImpl.VIEWID_QUERYSTRING_ATTRIBUTE, viewId.substring(qsLoc));
    }
         
    return viewRoot;

  }
  
  private String setUIViewRootComponent(FacesContext context)
  {
    Application app = context.getApplication();
    UIComponent root = app.createComponent(UIViewRoot.COMPONENT_TYPE);
    if (root == null || (root.getClass() == UIViewRoot.class && 
        root.getClass().getAnnotation(PortletNamingContainer.class) == null))
    {
      app.addComponent(UIViewRoot.COMPONENT_TYPE, PortletNamingContainerUIViewRoot.class.getName());
      return UIViewRoot.class.getName();
    }
    
    return null;
  }
  
  private void resetUIViewRootComponent(FacesContext context, String componentClassName)
  {
    if (componentClassName != null)
    {
      context.getApplication().addComponent(UIViewRoot.COMPONENT_TYPE, componentClassName);
    }
  }
  
  
  @Override
  public void renderView(FacesContext context, UIViewRoot viewToRender) throws IOException,
                                                                       FacesException
  {
    ExternalContext extContext = context.getExternalContext();
      
    // Do nothing when not running in portlet request
    if (!BridgeUtil.isPortletRequest())
    {
      super.renderView(context, viewToRender);
      return;
    }
    
    MimeResponse mimeResponse = (MimeResponse) extContext.getResponse();

    // If first time -- Get the renderPolicy from the context init parameter 
    if (mRenderPolicy == null)
    {
      PortletContext pCtx = (PortletContext) context.getExternalContext().getContext();
      String policy = pCtx.getInitParameter(Bridge.RENDER_POLICY);
      if (policy != null)
      {
        mRenderPolicy = Bridge.BridgeRenderPolicy.valueOf(policy);
      }
      else
      {
        mRenderPolicy = Bridge.BridgeRenderPolicy.DEFAULT;
      }
    }

    if (mRenderPolicy == Bridge.BridgeRenderPolicy.ALWAYS_DELEGATE)
    {
      super.renderView(context, viewToRender);
      return;
    }
    else if (mRenderPolicy == Bridge.BridgeRenderPolicy.DEFAULT)
    {
      try
      {
        super.renderView(context, viewToRender);
        return;
      }
      catch (Throwable t)
      {
        // catch all throws and swallow -- falling through to our own

        // render -- Note because delegate may have set the Response object and
        // not cleaned it up on the exception -- reset to ensure a good object
        if (!mimeResponse.equals(extContext.getResponse()))
        {
          extContext.setResponse(mimeResponse);
        }
        
        System.out.println(t);
      }
    }
    
    extContext.getRequestMap().remove("javax.servlet.include.servlet_path");
    getViewDeclarationLanguage(context,viewToRender.getViewId()).renderView(context, viewToRender);
  }
}
