/*
 * Licensed to the Apache Software Foundation (ASF) under one or more contributor license
 * agreements. See the NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The ASF licenses this file to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance with the License. You may obtain a
 * copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable
 * law or agreed to in writing, software distributed under the License is distributed on an "AS IS"
 * BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License
 * for the specific language governing permissions and limitations under the License.
 */

package org.apache.myfaces.portlet.faces.application.view;

import java.util.regex.Pattern;

import javax.faces.application.ViewHandler;
import javax.faces.context.ExternalContext;
import javax.faces.context.FacesContext;
import javax.faces.view.ViewDeclarationLanguage;
import javax.faces.view.ViewDeclarationLanguageFactory;

import javax.portlet.PortletContext;
import javax.portlet.faces.Bridge;
import javax.portlet.faces.BridgeUtil;

import org.apache.myfaces.buildtools.maven2.plugin.builder.annotation.JSFWebConfigParam;

/**
 * This is an implementation of the ViewDeclarationLanguageFactory for the portal.  This implementation first checks to see if
 * we have a facelet view.  If we do, we simply provide the ViewDeclarationLanguage object from the JSF Implementation.  If we
 * do NOT have a facelet, then the factory will provide an ViewDeclarationLanguage object that is appropriate to run JSP's
 * according to the Bridge.RENDER_POLICY.
 * 
 * @author sobryan
 */
public class PortletViewDeclarationLanguageFactoryImpl
  extends ViewDeclarationLanguageFactory
{
  /**
   * Disable facelets VDL from the current application project. 
   */
  @JSFWebConfigParam(since="2.0", defaultValue="false", expectedValues="true,false")
  public static final String PARAM_DISABLE_JSF_FACELET = "javax.faces.DISABLE_FACELET_JSF_VIEWHANDLER";
  
  private ViewDeclarationLanguageFactory mWrapped;
  private ViewDeclarationLanguage mVdl;
  private Pattern mAcceptPatterns;
  private String mExtension;
  
  public PortletViewDeclarationLanguageFactoryImpl(ViewDeclarationLanguageFactory wrapped)
  {
    mWrapped = wrapped;
    
    ExternalContext ec = FacesContext.getCurrentInstance().getExternalContext();

    mAcceptPatterns = loadAcceptPattern(ec);
    mExtension = loadFaceletExtension(ec);
  }

  public ViewDeclarationLanguage getViewDeclarationLanguage(String viewId)
  {
    ExternalContext ec = FacesContext.getCurrentInstance().getExternalContext();
    ViewDeclarationLanguage vdl = getWrapped().getViewDeclarationLanguage(viewId);
   
    if (BridgeUtil.isPortletRequest() && !isFacelet(ec, viewId))
    {
      // If first time -- Get the renderPolicy from the context init parameter 
      if (null == mVdl)
      {
        PortletContext pCtx = (PortletContext) ec.getContext();
        String policy = pCtx.getInitParameter(Bridge.RENDER_POLICY);
        
        if (null == policy)
        {
          policy = "DEFAULT";
        }
               
        switch(Bridge.BridgeRenderPolicy.valueOf(policy))
        {
          case ALWAYS_DELEGATE:
            mVdl = vdl;
            break;
          case DEFAULT:
            mVdl = new AggregatingViewDeclatationLanguage(vdl, new PortletJspViewDeclarationLanguageImpl());
            break;
          case NEVER_DELEGATE:
            mVdl = new PortletJspViewDeclarationLanguageImpl();
        }
      }
        
      return mVdl;
    }
      
    return vdl;
  }

  @Override
  public ViewDeclarationLanguageFactory getWrapped()
  {
    return mWrapped;
  }
  
  private boolean isFacelet(ExternalContext ec, String viewId)
  {
    String param = ec.getInitParameter(PARAM_DISABLE_JSF_FACELET);
    if(null != param && Boolean.parseBoolean(param.toLowerCase()))
    {
      //Facelet support is disabled so this will never be a facelet
      return false;
    }
    
    // Check extension first as it's faster than mappings
    if (viewId.endsWith(mExtension))
    {
        // If the extension matches, it's a Facelet viewId.
        return true;
    }

    // Otherwise, try to match the view identifier with the facelet mappings
    return mAcceptPatterns != null && mAcceptPatterns.matcher(viewId).matches();

  }
  
  /**
   * Load and compile a regular expression pattern built from the Facelet view mapping parameters.
   * 
   * @param context
   *            the application's external context
   * 
   * @return the compiled regular expression
   */
  //TODO: Expose these in Shared.  They come from the MyFaces FaceletViewDeclarationStrategy
  private Pattern loadAcceptPattern(ExternalContext context)
  {
      assert context != null;

      String mappings = context.getInitParameter(ViewHandler.FACELETS_VIEW_MAPPINGS_PARAM_NAME);
      if (mappings == null)
      {
          return null;
      }

      // Make sure the mappings contain something
      mappings = mappings.trim();
      if (mappings.length() == 0)
      {
          return null;
      }

      return Pattern.compile(toRegex(mappings));
  }

  //TODO: Expose these in Shared.  They come from the MyFaces FaceletViewDeclarationStrategy
  private String loadFaceletExtension(ExternalContext context)
  {
      assert context != null;

      String suffix = context.getInitParameter(ViewHandler.FACELETS_SUFFIX_PARAM_NAME);
      if (suffix == null)
      {
          suffix = ViewHandler.DEFAULT_FACELETS_SUFFIX;
      }
      else
      {
          suffix = suffix.trim();
          if (suffix.length() == 0)
          {
              suffix = ViewHandler.DEFAULT_FACELETS_SUFFIX;
          }
      }

      return suffix;
  }

  /**
   * Convert the specified mapping string to an equivalent regular expression.
   * 
   * @param mappings
   *            le mapping string
   * 
   * @return an uncompiled regular expression representing the mappings
   */
  //TODO: Expose these in Shared.  They come from the MyFaces FaceletViewDeclarationStrategy
  private String toRegex(String mappings)
  {
      assert mappings != null;

      // Get rid of spaces
      mappings = mappings.replaceAll("\\s", "");

      // Escape '.'
      mappings = mappings.replaceAll("\\.", "\\\\.");

      // Change '*' to '.*' to represent any match
      mappings = mappings.replaceAll("\\*", ".*");

      // Split the mappings by changing ';' to '|'
      mappings = mappings.replaceAll(";", "|");

      return mappings;
  }
}
